/*
  uart2spi.c
  David Rowe
  3 March 2009

  Atmel AVR Atmega8L program that converts RS232 UART samples to TDM
  bus samples and implements a FIFO buffer between the two ports.

  RS232 samples are received from the router UART, buffered, them,
  then sent to the SPI port.  The SPI data is sent to the FXS Module
  DTX line in TDM bus format.

  The SPI port operates in slave mode, each SPI transation starts with
  a nSS pulse generated from the CPLD at a rate of 8 kHz.  UART
  samples arrive in bursts from the Interrupt Service Routine (ISR)
  running on the router.

  See ../src/README.txt for more information on the FXS interface
  design.
*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include "iocompat.h"

#define BUF_SZ 16

/* used for testing - creates a sinusoidal tone */
static char digital_milliwatt[] = {0x1e,0x0b,0x0b,0x1e,0x9e,0x8b,0x8b,0x9e} ;

int main (void)
{
    unsigned char rx, *pin, *pout, count;
    int      nbuf;
    unsigned char buf[BUF_SZ];
	
    /* Set up SPI as slave -----------------------------*/

    DDRB |= (1<<PB4);  /* enable MISO as an output */
    SPCR = (1<<SPE) | (1<<CPOL); 

    /* configure output pins ---------------------------*/

    DDRB  |= (1<<PB0) | (1<<PB1);
	
    /* set up RS232 UART -------------------------------*/

    // set baud rate

    UBRRH = 0;
    UBRRL = 3;            // 115,200 baud with 7.3728 MHz clock

    // Enable receiver

    UCSRB = (1 << RXEN);

    // Set frame format: 8 data bits, 1 stop bit
    // Note URSEL is set so we can write to UCSRC which shares 
    // the same address as UBRRH above, state of URSEL selects
    // which register we write to.

    UCSRC = (1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0);

    /* set up buffer -----------------------------------*/

    nbuf = 0;
    pin = pout = buf;
    count = 0;

    while(1) {

	/* send byte via SPI port */

	if (SPSR & (1<<SPIF)) {  /* if previous SPI transfer complete */
	    rx = SPDR;
	    if (nbuf) {
		SPDR = *pout++;
		nbuf--;
	    }
	    else
		SPDR = 0x55; /* send 0 if FIFO empty */

	    if (pout == (buf + BUF_SZ))
		pout = buf;
		
            /* blink PB1 */

            PORTB |= 1 << PB1;
            PORTB &= ~(1<<PB1);
	}

        /* receive byte via RS232 UART */

        if (UCSRA & (1<<RXC)) {
	    if (nbuf <= BUF_SZ) {
		*pin++ = UDR;
		if (pin == (buf + BUF_SZ))
		    pin = buf;
		nbuf++;
	    }

	    /* blink PB0 */

            PORTB |= 1 << PB0;
            PORTB &= ~(1<<PB0);
	}
        
    }

    return (0);
}

