// receive.v
// David Rowe 5 Feb 2009
//
// Receive side processing, much simpler than transmit so
// all in one file.  Converts an 8 bit byte from the TDM
// bus to a 10 bit RS232 UART word.  Note that TDM bus data
// is sent in MSB:LSB order, RS232 is sent in LSB:MSB order.
//
// FXS chipset -> receive.v -> Router UART

module receive (rst, clk, drx, fs, rx);
input        rst;
input        clk;           // 2.048MHz system clock
input        drx;           // TDM bus input from FXS TDM output
input        fs;            // TDM bus frame clock
output       rx;            // RS232 rx data to router
reg          rx;

reg  [2:0]   state;
reg  [7:0]   data;
reg  [4:0]   clkdiv;        // clock divider to generate 115000 baud
reg  [2:0]   nbits;         // bits received so far

parameter IDLE   = 3'b000;  // waiting for next byte from TDM bus 
parameter SAMPLE = 3'b001;  // shifting in bits from TDM bus
parameter START  = 3'b010;  // send RS232 start bit 
parameter SEND   = 3'b011;  // shift out 8 data bits
parameter STOP   = 3'b100;  // send RS232 stop bits

always @(negedge clk)
begin
  if (rst) begin
      state       <= 0;
      rx          <= 1;
  end else begin

    case(state)

    IDLE:   if (fs) begin
              state  <= #1 SAMPLE;
              nbits  <= 0;
	      data   <= {data[6:0], drx}; // TDM is MSB first
            end
    
    SAMPLE: begin
              if (nbits == 7) begin
                state  <= #1 START;
                rx     <= 0;
                clkdiv <= 0;
              end
              else begin
                nbits <= nbits + 1;
	        data  <= {data[6:0], drx}; // TDM is MSB first
              end 
            end
   
    START:  begin
              clkdiv <= clkdiv + 1;
            
              if (clkdiv == 17) begin
                state  <= #1 SEND;
                rx     <= data[0];
                data   <= {1'b0, data[7:1]}; // RS232 is LSB first
                clkdiv <= 0;
                nbits  <= 0;
              end
            end

    SEND:   begin
              clkdiv <= clkdiv + 1;
            
              if (clkdiv == 17) begin
                nbits  <= nbits + 1;
                clkdiv <= 0;
                if (nbits == 7) begin
                  state  <= #1 STOP;
                  rx     <= 1;
                end
                else begin
                  rx     <= data[0];
                  data   <= {1'b0, data[7:1]}; // RS232 is LSB first
                end
              end
            end

    STOP:  begin
              clkdiv <= clkdiv + 1;
            
              if (clkdiv == 17) begin
                state  <= #1 IDLE;
              end
            end

    default: state <= IDLE;

    endcase
  end

end

endmodule
