// sfifo.v
// David Rowe Feb 4 2009
//
// Serial input, serial output, depth is set by ADDR_WIDTH below
//
// Reading:
//   + take rd high
//   + clock in 8 data bits
//   + take rd low, this increments read pointer on next clock
//
// Writing:
//   + take wr high
//   + clock out 8 data bits
//   + take wr low, this increments read pointer on next clock
//
// Note: we allow simultaneous writing and reading.  We just hope that
//       it doesn't happen to the same memory location at the same time
//       as this would be an over/underflow situation.

module sfifo (rst, clk, wr, wr_finished, din, rd, rd_finished, dout);
input        rst;
input        clk;
input        wr;          // assert when writing an 8 bit serial word
input        wr_finished; // assert after 8 bits have been written
input        din;         // serial input data
input        rd;          // assert when reading an 8 bit serial word
input        rd_finished; // assert after 8 bits have been read
output       dout;        // serial output data

parameter DATA_WIDTH = 8;
parameter ADDR_WIDTH = 1;
parameter RAM_DEPTH = 1 << ADDR_WIDTH;

wire                  dout;
reg [DATA_WIDTH-1:0]  mem [0:RAM_DEPTH-1]; // FIFO memory
reg [ADDR_WIDTH-1:0]  pwr;                 // write pointer
reg [ADDR_WIDTH-1:0]  prd;                 // read pointer
wire [DATA_WIDTH-1:0] wrdata;
wire [DATA_WIDTH-1:0] new_wrdata;
wire [DATA_WIDTH-1:0] rddata;
wire [DATA_WIDTH-1:0] new_rddata;

  always @(posedge clk)
  begin

    if (rst) begin
      pwr <= 0;
      prd <= 0;
    end

    if (wr) begin
      mem[pwr] <= new_wrdata;
    end

    if (rd) begin
      mem[prd] <= new_rddata;
    end

    if (wr_finished) begin
      pwr <= pwr + 1;
    end	

    if (rd_finished) begin
      prd <= prd + 1;
    end	
  end

  assign dout = rddata[7];
  assign wrdata = mem[pwr];
  assign new_wrdata = { wrdata[6:0], din };
  assign rddata = mem[prd];
  assign new_rddata = { rddata[6:0], 1'b0 };
endmodule
